
class ParamError(Exception):
    pass


def read_param(request, param, default=None, type=str):
    try:
        value = request.args.get(param, request.form.get(param))
        if value is None:
            return default
        else:
            return type(value)
    except (ValueError, TypeError) as e:
        raise ParamError("Could not interpret {0}. {1}"
                         .format(param, str(e)))


def read_bar_split_param(request, param, default=None, type=str):
    values = read_param(request, param, default=default)
    if values is None:
        return []

    try:
        return [type(value) for value in values.split("|")]
    except (ValueError, TypeError) as e:
        raise ParamError("Could not interpret {0}. {1}"
                         .format(param, str(e)))



"""
`revscoring` is a generic, revision scoring system designed to be used to
automatically "score" revisions in a MediaWiki installation.  This library
eases the construction and use of Machine Learning model-based scoring
strategies.

* See the :ref:`API reference <api-reference>` for detailed information

Key Features
------------

Scorer Models:
    :class:`~revscoring.scorer_models.scorer_model.ScorerModel` are the core of
    the `revscoring` system.  Provide a simple interface with complex
    internals.  Most commonly, :class:`~revscoring.scorer_models.MLScorerModel`
    (Machine Learned) is
    :meth:`~revscoring.scorer_models.MLScorerModel.train`'d and
    :meth:`~revscoring.scorer_models.MLScorerModel.test`'d on
    labeled data to provide a basis for scoring.
    We currently support: :mod:`~revscoring.scorer_models.svc`,
    :mod:`~revscoring.scorer_models.rf` and :mod:`~revscoring.scorer_models.nb`
    type models. (see :mod:`revscoring.scorer_models`)

    Example: *TODO*

Feature extraction:
    Revscoring provides a dependency-injection-based feature extraction
    framework that allows new features to be built on top of old.  This allows
    a powerful means to expressing new features and a simple way to address
    efficiency concerns. (see :mod:`revscoring.features`,
    :mod:`revscoring.datasources`, and :mod:`revscoring.extractors`)

    Example: *TODO*

Language support:
    Many features require language specific utilities to be available to
    support feature extraction.  In order to support this, we provide a
    collection of language utilities that integrate with our feature
    extraction system so as to be requested when necessary.  We support
    language utilities for the following languages:
    :data:`~revscoring.languages.english`,
    :data:`~revscoring.languages.french`,
    :data:`~revscoring.languages.hebrew`,
    :data:`~revscoring.languages.indonesian`,
    :data:`~revscoring.languages.persian`,
    :data:`~revscoring.languages.portuguese`,
    :data:`~revscoring.languages.spanish`,
    :data:`~revscoring.languages.turkish`, and
    :data:`~revscoring.languages.vietnamese`.
    (see :mod:`revscoring.languages`)

    Example: *TODO*
"""
__version__ = "0.6.5"



"""
This module implements a set of :class:`~revscoring.language.Language`
-- collections of features that are language specific.

languages
+++++++++
.. automodule:: revscoring.languages.english

.. automodule:: revscoring.languages.french

.. automodule:: revscoring.languages.hebrew

.. automodule:: revscoring.languages.indonesian

.. automodule:: revscoring.languages.persian

.. automodule:: revscoring.languages.portuguese

.. automodule:: revscoring.languages.spanish
    :members:

.. automodule:: revscoring.languages.turkish
    :members:

.. automodule:: revscoring.languages.vietnamese
    :members:

Base classes
++++++++++++
.. automodule:: revscoring.languages.language

.. automodule:: revscoring.languages.space_delimited

"""
from .language import Language

__all__ = [Language]



import pickle

from nose.tools import eq_

from .. import french
from ...datasources import revision
from ...dependencies import solve

BAD = [
    "con",
    "fesse",
    "foutre",
    "merde", "merdee",
    "merdique",
    "prostituee", "prostitue",
    "putain", "putes",
    "salop",
    "stupide"
]

OTHER = [
    "connection", "fitness", "le"
]


def compare_extraction(extractor, examples, counter_examples):

    for example in examples:
        eq_(extractor.process(example), [example])
        eq_(extractor.process("Sentence " + example + " sandwich."), [example])
        eq_(extractor.process("Sentence end " + example + "."), [example])
        eq_(extractor.process(example + " start of sentence."), [example])

    for example in counter_examples:
        eq_(extractor.process(example), [])
        eq_(extractor.process("Sentence " + example + " sandwich."), [])
        eq_(extractor.process("Sentence end " + example + "."), [])
        eq_(extractor.process(example + " start of sentence."), [])


def test_badwords():
    compare_extraction(french.revision.badwords_list, BAD, OTHER)


def test_revision():
    # Words
    cache = {revision.text: "Wikipédia est un projet d’encyclopédie."}
    eq_(solve(french.revision.words_list, cache=cache),
        ["Wikipédia", "est", "un", "projet", "d’encyclopédie"])

    # Misspellings
    cache = {revision.text: 'Est un projet principe du worngly. <td>'}
    eq_(solve(french.revision.misspellings_list, cache=cache), ["worngly"])

    # Infonoise
    cache = {revision.text: "Est un projet principe."}
    eq_(solve(french.revision.infonoise, cache=cache), 13/19)


def test_presence():
    assert hasattr(french.revision, "words")
    assert hasattr(french.revision, "content_words")
    assert hasattr(french.revision, "badwords")
    assert hasattr(french.revision, "misspellings")

    assert hasattr(french.parent_revision, "words")
    assert hasattr(french.parent_revision, "content_words")
    assert hasattr(french.parent_revision, "badwords")
    assert hasattr(french.parent_revision, "misspellings")

    assert hasattr(french.diff, "words_added")
    assert hasattr(french.diff, "badwords_added")
    assert hasattr(french.diff, "misspellings_added")
    assert hasattr(french.diff, "words_removed")
    assert hasattr(french.diff, "badwords_removed")
    assert hasattr(french.diff, "misspellings_removed")


def test_pickling():

    eq_(french, pickle.loads(pickle.dumps(french)))



"""
``revscoring extract_features -h``
::

    Adds features to a set of labeled revisions.

    Reads a TSV file of <rev_id>\t<label> pairs and replaces the
    <rev_id> field with the extracted feature values.

    Input: <rev_id>[TAB]<label>

    Output: <feature_1>[TAB]<feature_2>[TAB]...[TAB]<label>


    Usage:
        extract_features -h | --help
        extract_features <features> --host=<url> [--rev-labels=<path>]
                                                 [--value-labels=<path>]
                                                 [--verbose] [--debug]

    Options:
        -h --help                Print this documentation
        <features>               Classpath to a list/tuple of features
        --host=<url>             The url pointing to a MediaWiki API to use
                                 for extracting features
        --rev-labels=<path>      Path to a file containing rev_id-label pairs
                                 [default: <stdin>]
        --value-labels=<path>    Path to a file to write feature-labels to
                                 [default: <stdout>]
        --verbose                Print dots and stuff
        --debug                  Print debug logging
"""
import logging
import sys
from itertools import islice

import docopt
import mwapi

from ..errors import RevisionNotFound
from ..extractors import APIExtractor
from .util import encode, import_from_path

logger = logging.getLogger(__name__)

BATCH_SIZE = 50


def main(argv=None):
    args = docopt.docopt(__doc__, argv=argv)

    features = import_from_path(args['<features>'])

    session = mwapi.Session(args['--host'],
                            user_agent="Revscoring feature extractor utility")
    extractor = APIExtractor(session)

    if args['--rev-labels'] == "<stdin>":
        rev_labels = read_rev_labels(sys.stdin)
    else:
        rev_labels = read_rev_labels(open(args['--rev-labels']))

    if args['--value-labels'] == "<stdout>":
        value_labels = sys.stdout
    else:
        value_labels = open(args['--value-labels'], 'w')

    verbose = args['--verbose']
    debug = args['--debug']

    run(rev_labels, value_labels, features, extractor, verbose, debug)


def read_rev_labels(f):
    # Check if first line is a header
    rev_id, label = f.readline().strip().split("\t")
    if rev_id != "rev_id":
        yield int(rev_id), label

    for line in f:
        rev_id, label = line.strip().split('\t')
        yield int(rev_id), label


def run(rev_labels, value_labels, features, extractor, verbose, debug):
    logging.basicConfig(
        level=logging.WARNING if not debug else logging.DEBUG,
        format='%(asctime)s %(levelname)s:%(name)s -- %(message)s'
    )

    while True:
        batch_rev_labels = list(islice(rev_labels, BATCH_SIZE))
        if len(batch_rev_labels) == 0:
            break

        rev_ids, labels = zip(*batch_rev_labels)

        error_values_label = zip(extractor.extract(rev_ids, features),
                                 batch_rev_labels)
        for (error, values), (rev_id, label) in error_values_label:
            try:
                if isinstance(error, RevisionNotFound):
                    if verbose:
                        sys.stderr.write("?")
                        sys.stderr.flush()
                elif error is not None:
                    logger.error("An error occured while processing {0}:"
                                 .format(rev_id))
                    logger.error("\t{0}: {1}"
                                 .format(error.__class__.__name__,
                                         str(error)))
                else:
                    fields = list(values) + [label]
                    value_labels.write("\t".join(encode(v)
                                                 for v in fields))
                    value_labels.write("\n")

                    if verbose:
                        sys.stderr.write(".")
                        sys.stderr.flush()

            except KeyboardInterrupt:
                sys.stderr.write("^C detected.  Shutting down.\n")
                break

    if verbose:
        sys.stderr.write("\n")



def configure(config):

    from flask import Blueprint, Flask

    from . import routes
    from ..score_processors import ScoreProcessor

    app = Flask(__name__)
    app.config['APPLICATION_ROOT'] = config['ores']['wsgi']['application_root']

    bp = Blueprint('ores', __name__)

    sp_name = config['ores']['score_processor']
    score_processor = ScoreProcessor.from_config(config, sp_name)

    bp = routes.configure(config, bp, score_processor)

    app.register_blueprint(bp, url_prefix=config['ores']['wsgi']['url_prefix'])

    return app



import sys
from importlib import import_module

sys.path.insert(0, ".")  # Necessary for working in other modules


def import_from_path(path):
    try:
        module = import_module(path)
        return module
    except ImportError:
        parts = path.split(".")
        module_path = ".".join(parts[:-1])
        attribute_name = parts[-1]

        module = import_module(module_path)

        attribute = getattr(module, attribute_name)

        return attribute


def encode(val, none_val="NULL"):
    if val is None:
        return none_val
    elif isinstance(val, bytes):
        val = str(val, 'utf-8', "replace")
    else:
        val = str(val)

    return val.replace("\t", "\\t").replace("\n", "\\n")
